part of katana_routing;

/// Extended method for Navigator.
extension NavigatorStateExtensions on NavigatorState {
  /// Pops the page to the page with the specified path.
  ///
  /// Enter the pathname of the page in [name].
  Future<void> popUntilNamed(String name) async {
    popUntil((route) {
      return route.settings.name == name;
    });
  }

  /// Reset all history and go to a specific page.
  ///
  /// After all pages are [pop()],
  /// the page specified by [newRouteName] is [pushNamed()].
  ///
  /// In [arguments], put [RouteQuery] and so on.
  ///
  /// Ideal for moving after logging out.
  Future<T?> resetAndPushNamed<T extends Object?>(
    String newRouteName, {
    Object? arguments,
  }) {
    return pushNamedAndRemoveUntil(
      newRouteName,
      (route) => false,
      arguments: arguments,
    );
  }
}

extension BuildContextNavigatorStateExtensions
    on _BuildContextNavigatorContainer {
  /// Pops the page to the page with the specified path.
  ///
  /// Enter the pathname of the page in [name].
  Future<void> popUntilNamed(String name) async {
    final parsedName = _parse(name);
    popUntil((route) {
      return _parse(route.settings.name ?? "") == parsedName;
    });
  }

  /// Reset all history and go to a specific page.
  ///
  /// After all pages are [pop()],
  /// the page specified by [newRouteName] is [pushNamed()].
  ///
  /// In [arguments], put [RouteQuery] and so on.
  ///
  /// Ideal for moving after logging out.
  Future<T?> resetAndPushNamed<T extends Object?>(
    String newRouteName, {
    Object? arguments,
  }) {
    return pushNamedAndRemoveUntil(
      newRouteName,
      (route) => false,
      arguments: arguments,
    );
  }
}

/// Extension methods for BuildContext.
extension BuildContedxtExtensions on BuildContext {
  DynamicMap get _arg {
    final pageScoped = PageScoped._of(this);
    return pageScoped._modalRoute?.settings.arguments as DynamicMap? ??
        pageScoped._map;
  }

  /// Outputs the theme related to the context.
  ThemeData get theme => Theme.of(this);

  /// Outputs the widget theme related to the context.
  WidgetTheme get widgetTheme => WidgetTheme.of(this);

  /// Outputs the image theme related to the context.
  ImageTheme get imageTheme => ImageTheme.of(this);

  /// Get the Navigator related to context.
  NavigatorState get navigator {
    final navigator = Navigator.of(this);
    return _BuildContextNavigatorContainer(navigator, this);
  }

  /// Get the Root navigator related to context.
  NavigatorState get rootNavigator {
    final navigator = Navigator.of(this, rootNavigator: true);
    return _BuildContextNavigatorContainer(navigator, this);
  }

  /// Get the Flavor.
  String get flavor => FlavorScope.of(this).flavor;

  /// Get the Design type.
  DesignType get designType => DesignTypeScope.of(this).designType;

  /// If you want to use the style for web when you are on the web, use `true`.
  bool get webStyle => DesignTypeScope.of(this).webStyle;

  /// Get the route settins.
  RouteSettings get route =>
      ModalRoute.of(this)?.settings ?? const RouteSettings();

  /// Get the media qury related to context
  MediaQueryData get mediaQuery => MediaQuery.of(this);

  /// Releases focus such as text field.
  void unfocus() {
    FocusManager.instance.primaryFocus?.unfocus();
  }

  /// True if the specified [key] is contained in scoped.
  bool containsKey(Object? key) {
    return _arg.containsKey(key);
  }

  /// True if the specified [value] is contained in scoped.
  bool containsValue(Object? value) {
    return _arg.containsValue(value);
  }

  /// Gets the value from the [key] of the map stored in scoped.
  ///
  /// If [key] does not exist in the map, [orElse] is returned.
  T get<T>(String key, T orElse) {
    return _arg.get(key, orElse);
  }

  /// Get the list corresponding to [key] in the map.
  ///
  /// If [key] is not found, the list of [orElse] is returned.
  List<T> getAsList<T>(String key, [List<T>? orElse]) {
    if (!_arg.containsKey(key)) {
      return orElse ?? [];
    }
    return (_arg[key] as List?)?.cast<T>() ?? orElse ?? [];
  }

  /// Get the map corresponding to [key] in the map.
  ///
  /// If [key] is not found, the map of [orElse] is returned.
  Map<String, T> getAsMap<T>(String key, [Map<String, T>? orElse]) {
    if (!_arg.containsKey(key)) {
      return orElse ?? {};
    }
    return (_arg[key] as Map?)?.cast<String, T>() ?? orElse ?? {};
  }

  /// Get the set corresponding to [key] in the map.
  ///
  /// If [key] is not found, the set of [orElse] is returned.
  Set<T> getAsSets<T>(String key, [Set<T>? orElse]) {
    if (!_arg.containsKey(key)) {
      return orElse ?? {};
    }
    return (_arg[key] as Set?)?.cast<T>() ?? orElse ?? {};
  }

  /// Get the set corresponding to [key] in the DateTime.
  ///
  /// If [key] is not found, the set of [orElse] is returned.
  DateTime getAsDateTime(String key, [DateTime? orElse]) {
    if (!_arg.containsKey(key)) {
      return orElse ?? DateTime.now();
    }
    final millisecondsSinceEpoch = _arg[key] as int?;
    if (millisecondsSinceEpoch == null) {
      return orElse ?? DateTime.now();
    }
    return DateTime.fromMillisecondsSinceEpoch(millisecondsSinceEpoch);
  }

  /// Get [value] to [key] of the map stored in [arg].
  ///
  /// The data will be unique within the page and can continue to be used even if the widget is rebuilt.
  dynamic operator [](Object? key) {
    return _arg[key];
  }

  /// Save [value] to [key] of the map stored in [arg].
  ///
  /// The data will be unique within the page and can continue to be used even if the widget is rebuilt.
  void operator []=(String key, dynamic value) {
    _arg[key] = value;
  }
}

/// Extension method for WidgetRef.
extension WidgetRefExtensions on WidgetRef {
  DynamicMap get _arg {
    final scoped = Scoped.of(this as BuildContext);
    return scoped._map;
  }

  /// Explicitly rebuild the widget.
  ///
  /// Basically, do not use it.
  ///
  /// Use this only if you want to force a rebuild.
  void rebuild() {
    Scoped.of(this as BuildContext).rebuild();
  }

  /// Open a new URL or screen.
  ///
  /// If [path] starts with `http`,
  /// it opens a browser, otherwise it opens a screen path.
  ///
  /// It is possible to configure the settings for
  /// opening the screen by specifying [query].
  Future<T?> open<T extends Object?>(String path, [RouteQuery? query]) async {
    if (path.startsWith("http")) {
      await openURL(path);
      return null;
    } else {
      return await (this as BuildContext)
          .navigator
          .pushNamed(path, arguments: query);
    }
  }

  /// A builder for creating and saving scoped value.
  ///
  /// If [key] is specified, the data will be saved with that key.
  ///
  /// Since it is not usually used directly from [get], [prefix] is given to prevent duplication of values.
  ///
  /// It creates a scoped value from [builder].
  ///
  /// If [test] is `true` or the scoped value for the Key is empty, update the value.
  @protected
  TResult valueBuilder<TResult, TScopedValue extends ScopedValue<TResult>>({
    required String key,
    required TScopedValue Function() builder,
    String prefix = r"_$",
  }) {
    key = "$prefix$key";
    final found = _arg.get<ScopedValueState<TResult, TScopedValue>?>(key, null);
    final value = builder.call();
    if (found != null) {
      final oldValue = found.value;
      found._ref = this;
      found._value = value;
      found.didUpdateValue(oldValue);
      return found.build();
    } else {
      final state = value.createState();
      state._ref = this;
      state._value = value;
      state.initValue();
      _arg[key] = state;
      return state.build();
    }
  }
}

/// Gradient color extension for Routing.
extension GradientColorExtensions on Color {
  /// Gets the color for the gradient.
  ///
  /// If GradientColor is specified in Theme, an array of it can be obtained.
  List<Color> toGradient() {
    final color = this;
    if (color is GradientColor) {
      return color._toGradient();
    } else {
      return [this];
    }
  }

  /// True if the color is a gradient color.
  bool get isGradient {
    return this is GradientColor;
  }
}

/// TextStyle extension for Routing.
extension RoutingTextStyleExtensions on TextStyle {
  /// Makes the text color darker.
  ///
  /// The higher the value(0.0 - 1.0), the darker the image becomes.
  TextStyle darken([double amount = 0.1]) {
    return copyWith(
      color: color?.darken(amount),
    );
  }

  /// Makes the text color lighter.
  ///
  /// The higher the value(0.0 - 1.0), the lighter the image becomes.
  TextStyle lighten([double amount = 0.1]) {
    return copyWith(
      color: color?.lighten(amount),
    );
  }

  /// Set the transparency of the text color to [opacity].
  TextStyle withOpacity(double opacity) {
    return copyWith(color: color?.withOpacity(opacity));
  }

  /// Decrease the font size of TextStyle by [amount].
  TextStyle smallize([double amount = 1]) {
    return copyWith(fontSize: fontSize == null ? null : (fontSize! - amount));
  }

  /// Increase the font size of TextStyle by [amount].
  TextStyle largize([double amount = 1]) {
    return copyWith(fontSize: fontSize == null ? null : (fontSize! + amount));
  }
}

/// Theme extension for Routing.
extension RoutingThemeExtension on ThemeData {
  /// Text color.
  Color get textColor {
    return colorScheme.onBackground;
  }

  /// Text color on primary.
  Color get textColorOnPrimary {
    return colorScheme.onPrimary;
  }

  /// Text color on secondary.
  Color get textColorOnSecondary {
    return colorScheme.onSecondary;
  }

  /// Text color on tertiary.
  Color get textColorOnTertiary {
    return colorScheme.onTertiary;
  }

  /// Text color on surface.
  Color get textColorOnSurface {
    return colorScheme.onSurface;
  }

  /// Text color on error.
  Color get textColorOnError {
    return colorScheme.onError;
  }

  /// Text color on warning.
  Color get textColorOnWarning {
    return colorScheme.onError;
  }

  /// Secondary color.
  Color get secondaryColor {
    return colorScheme.secondary;
  }

  /// Tertiary color.
  Color get tertiaryColor {
    return colorScheme.tertiary;
  }

  /// Surface color.
  Color get surfaceColor {
    return colorScheme.surface;
  }

  /// Warning color.
  Color get warningColor {
    return colorScheme.errorContainer;
  }

  /// Display large text style.
  TextStyle get textDisplayLarge {
    return textTheme.displayLarge ??
        const TextStyle(
          fontSize: 57,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.12,
        );
  }

  /// Display medium text style.
  TextStyle get textDisplayMedium {
    return textTheme.displayMedium ??
        const TextStyle(
          fontSize: 45,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.15,
        );
  }

  /// Display small text style.
  TextStyle get textDisplaySmall {
    return textTheme.displaySmall ??
        const TextStyle(
          fontSize: 36,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.22,
        );
  }

  /// Headline large text style.
  TextStyle get textHeadlineLarge {
    return textTheme.headlineLarge ??
        const TextStyle(
          fontSize: 32,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.proportional,
          height: 1.25,
        );
  }

  /// Headline medium text style.
  TextStyle get textHeadlineMedium {
    return textTheme.headlineMedium ??
        const TextStyle(
          fontSize: 28,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.proportional,
          height: 1.29,
        );
  }

  /// Headline small text style.
  TextStyle get textHeadlineSmall {
    return textTheme.headlineSmall ??
        const TextStyle(
          fontSize: 24,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.proportional,
          height: 1.33,
        );
  }

  /// Title large text style.
  TextStyle get textTitleLarge {
    return textTheme.titleLarge ??
        const TextStyle(
          fontSize: 22,
          fontWeight: FontWeight.w500,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.27,
        );
  }

  /// Title medium text style.
  TextStyle get textTitleMedium {
    return textTheme.titleMedium ??
        const TextStyle(
          fontSize: 16,
          fontWeight: FontWeight.w500,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.5,
        );
  }

  /// Title small text style.
  TextStyle get textTitleSmall {
    return textTheme.titleSmall ??
        const TextStyle(
          fontSize: 14,
          fontWeight: FontWeight.w500,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.43,
        );
  }

  /// Body large text style.
  TextStyle get textBodyLarge {
    return textTheme.bodyLarge ??
        const TextStyle(
          fontSize: 16,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.5,
        );
  }

  /// Body medium text style.
  TextStyle get textBodyMedium {
    return textTheme.bodyMedium ??
        const TextStyle(
          fontSize: 14,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.43,
        );
  }

  /// Body small text style.
  TextStyle get textBodySmall {
    return textTheme.bodySmall ??
        const TextStyle(
          fontSize: 12,
          fontWeight: FontWeight.w400,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.33,
        );
  }

  /// Label large text style.
  TextStyle get textLabelLarge {
    return textTheme.labelLarge ??
        const TextStyle(
          fontSize: 14,
          fontWeight: FontWeight.w500,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.43,
        );
  }

  /// Label medium text style.
  TextStyle get textLabelMedium {
    return textTheme.labelMedium ??
        const TextStyle(
          fontSize: 12,
          fontWeight: FontWeight.w500,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.33,
        );
  }

  /// Label label text style.
  TextStyle get textLabelSmall {
    return textTheme.labelSmall ??
        const TextStyle(
          fontSize: 11,
          fontWeight: FontWeight.w500,
          leadingDistribution: TextLeadingDistribution.even,
          // height: 1.45,
        );
  }
}
